import java.util.*;
import java.io.*;

/**
 * <p>Title: K-Means Clustering</p>
 * <p>Description: Clustering using the K-Means approach</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * @author Huiyong Xiao
 * @version 1.0
 */

public class KMeans {
  private File namefile = null;
  private File datafile = null;

  private int k = -1; // k in 'KMeans'
  private ArrayList seedIDlist = new ArrayList(); // to store the IDs (NOs) of the tuples.
  private ArrayList centroidlist = new ArrayList(); // to store the centroids of the clusters.
  private ArrayList clusterlist = new ArrayList(); // to store the clusters. Each cluster is a list of IDs of the tuples
  private ArrayList attrlist = new ArrayList(); // to store the attributes
  private ArrayList valuelist = new ArrayList(); // to store all the tuples (cases). Each tuple is a list of attrvalues.

  public KMeans() {
  }

  public static void main(String[] args) {
    try {
      // load the file specified by filename.
      System.out.print("Please input the file name: ");
      BufferedReader lineReader = new BufferedReader(new InputStreamReader(
          System.in));
      String fileitemname = lineReader.readLine();

      KMeans km = new KMeans();
      if (km.loadFiles(fileitemname) == false) {
        return;
      }

      // handle the inputted 'K'.
      while (true) {
        System.out.print("Please input 'K' (int): ");
        try {
          int k = Integer.parseInt(lineReader.readLine());
          km.setK(k);
          break;
        }
        catch (NumberFormatException ee) {
          System.out.println("Format error! Please input an integer!!!");
        }
      }

      // handle the method of choosing seeds.
      while (true) {
        System.out.print("Pick up seeds Randomly or Specifically? (R/S): ");
        String st = lineReader.readLine();
        if (st.equals("R")) {
          km.seedingRandomly();
          break;
        }
        else if (st.equals("S")) {
          km.seedingSpecifically();
          break;
        }
        else {
          System.out.println("Input error! Please input 'R' or 'S'!!!");
        }
      }

      km.clustering();

      String outputFilename = "";
      do {
        System.out.print("Please specify the output file itemname: ");
        outputFilename = lineReader.readLine();
      }
      while (!km.output(outputFilename));

    }
    catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Validate file names and Load data from files.
   * @param fileitemname specifies the file name not including the extension.
   * @return TRUE: The files exist in the current directory;<br>
   * FALSE: Any of the files doesn't exist in the current directory.
   */
  public boolean loadFiles(String fileitemname) {
    // validate if the files exist.
    this.namefile = new File("./" + fileitemname + ".names");
    if (!this.namefile.exists()) {
      System.out.println("The file .names doesn't exist!");
      return false;
    }
    this.datafile = new File("./" + fileitemname + ".data");
    if (!this.datafile.exists()) {
      System.out.println("The file .data doesn't exist!");
      return false;
    }

    this.loadName();
    this.loadData();
    return true;
  }

  /**
   * Load attr from the '.names' file.
   */
  protected boolean loadName() {
    try {
      BufferedReader reader = new BufferedReader(new FileReader(this.namefile));
      String line = "";
      int ind = 0;
      while ( (line = reader.readLine()) != null) {
        // remove the annotation
        ind = line.indexOf('|');
        if (ind >= 0) {
          line = line.trim().substring(0, ind).trim();
        }
        if (line.equals("")) {
          continue;
        }

        ind = line.indexOf(':');

        if (ind >= 0) {
          // read attribute names into the "attrlist".
          String name = line.substring(0, ind);
          this.attrlist.add(name);
        }
      } // while readLine()

      return true;

    }
    catch (IOException ex) {
      System.out.println("File I/O error. Please verify the .names file");
      ex.printStackTrace();
      return false;
    }
  }

  /**
   * Load data from the '.data' file.
   */
  protected void loadData() {
    try {
      BufferedReader reader = new BufferedReader(new FileReader(this.datafile));
      String line = "";
      int ind = 0;
      while ( (line = reader.readLine()) != null) {
        // remove the annotation
        ind = line.indexOf('|');
        if (ind >= 0) {
          line = line.trim().substring(0, ind).trim();
        }
        if (line.equals("")) {
          continue;
        }

        ArrayList tuple = new ArrayList();
        ind = line.indexOf(',');
        while (ind > 0) {
          String attrvalue = line.substring(0, ind);
          tuple.add(new Double(attrvalue));

          line = line.substring(ind + 1).trim();
          ind = line.indexOf(',');
        } // while

        this.valuelist.add(tuple);

      } // while readLine()
    }
    catch (IOException ex) {
      System.out.println("File I/O error. Please verify the .data file");
      ex.printStackTrace();
    }
  }

  /**
   * Set the value of 'K'.
   */
  public void setK(int k) {
    this.k = k;
  }

  /**
   * Generate the k seeds randomly.
   */
  public void seedingRandomly() {
    int upper = this.valuelist.size();
    Random rg = new Random();
    for (int i = 0; i < k; i++) {
      this.seedIDlist.add(new Integer(rg.nextInt(upper))); // nextInt returns an integer between [0, uppper-1]
    }
  }

  /**
   * Generate the k seeds manually.
   */
  public void seedingSpecifically() {
    try {
      BufferedReader lineReader = new BufferedReader(new InputStreamReader(
          System.in));
      int upper = this.valuelist.size();
      int i = 1;
      while (i <= this.k) {
        System.out.print("Please input No " + i + " seed (int[0.." + (upper - 1) +
                         "]): ");
        try {
          int seedID = Integer.parseInt(lineReader.readLine());
          if (seedID < 0 || seedID > (upper - 1)) {
            System.out.println(
                "Format error! Please input an integer between [0.." +
                (upper - 1) +
                "]!!!");
            continue;
          }
          this.seedIDlist.add(new Integer(seedID));
          i++;
        }
        catch (NumberFormatException ee) {
          System.out.println(
              "Format error! Please input an integer between [0.." + (upper - 1) +
              "]!!!");
        }
      } // while
    }
    catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Cluster the whole dataset.
   * Inputs: 1. the tuples in the "valuelist".
   *         2. the seeds in the "seedIDlist". Also as the Centroids.
   * Output: 1. the clusters in the "clusterlist".
   */
  public void clustering() {
    // initialize the centroid list using the seedIDlist.
    for (int j = 0; j < this.k; j++) {
      int id = ( (Integer)this.seedIDlist.get(j)).intValue();
      this.centroidlist.add( (ArrayList)this.valuelist.get(id));
    }

    double oldError = 0;
    double newError = 0;
    double oldErrorDec = 0;
    double newErrorDec = 0;
    int loopNo = 0;
    do {
      // initialize the clusters in the cluster list
      this.clusterlist.clear();
      for (int i = 0; i < this.k; i++) {
        this.clusterlist.add(new ArrayList());
      }

      // clustering...
      for (int i = 0; i < this.valuelist.size(); i++) {
        ArrayList tuple = (ArrayList)this.valuelist.get(i);
        double minDistance = 0;
        int nearestCenterID = 0;

        // find the nearest centroid for the current tuple.
        for (int j = 0; j < this.k; j++) {
          ArrayList centroid = (ArrayList)this.centroidlist.get(j);

          // computer the Euclidean distance between the tuple and the centroid.
          double distance = euDistance(tuple, centroid);
          if (j == 0 || distance < minDistance) {
            minDistance = distance;
            nearestCenterID = j;
          }
        } // for j

        // add the Index (i) of the current tuple into the corresponding Cluster (nearestCenterID).
        ArrayList cluster = (ArrayList)this.clusterlist.get(nearestCenterID);
        cluster.add(new Integer(i));
        this.clusterlist.set(nearestCenterID, cluster);
      } // for i

      // recompute the centroids and update them in the 'centroidlist'.
      for (int j = 0; j < this.k; j++) {
        ArrayList newCentroid = this.mean( (ArrayList)this.clusterlist.get(j));
        this.centroidlist.set(j, newCentroid);
      }

      // check the loop-ending condition - squaredError.
      loopNo++;
      newError = this.squaredError();
      if (loopNo == 1) {
        oldError = newError;
      }
      else {
        newErrorDec = oldError - newError;
      }

      if (loopNo <= 2) {
        oldErrorDec = newErrorDec;
      }
      else if (oldErrorDec >= newErrorDec) {
        break;
      }
    }
    while (true);
  }

  /**
   * Compute the Euclidean Distance between two tuples.
   */
  protected double euDistance(ArrayList tuple1, ArrayList tuple2) {
    double sum = 0;
    for (int i = 0; i < tuple1.size(); i++) {
      double value1 = ( (Double) tuple1.get(i)).doubleValue();
      double value2 = ( (Double) tuple2.get(i)).doubleValue();

      sum += Math.pow( (value1 - value2), 2);
    }
    return Math.sqrt(sum);
  }

  /**
   * Compute the Squared Error.
   */
  protected double squaredError() {
    double error = 0;
    for (int i = 0; i < this.k; i++) {
      ArrayList cluster = (ArrayList)this.clusterlist.get(i);
      ArrayList centroid = (ArrayList)this.centroidlist.get(i); // the centroid is the mean of the cluster.

      double sum = 0;

      for (int j = 0; j < cluster.size(); j++) {
        ArrayList tuple = (ArrayList)this.valuelist.get( ( (Integer) cluster.
            get(i)).intValue());
        double distance = this.euDistance(centroid, tuple);
        sum += (distance * distance);
      }

      error += sum;
    }
    return error;
  }

  /**
   * Compute the mean of all the tuples in the each cluster.
   */
  protected ArrayList mean(ArrayList cluster) {
    ArrayList meanTuple = new ArrayList();

    // initialize each attrvalue to '0';
    for (int i = 0; i < this.attrlist.size(); i++) {
      meanTuple.add(new Double(0));
    }

    // accumulate.
    for (int i = 0; i < cluster.size(); i++) {
      ArrayList tuple = (ArrayList)this.valuelist.get( ( (Integer) cluster.get(
          i)).intValue());
      // accumulate each attrvalue.
      for (int j = 0; j < this.attrlist.size(); j++) {
        double sum = ( (Double) meanTuple.get(j)).doubleValue();
        double adder = ( (Double) tuple.get(j)).doubleValue();
        meanTuple.set(j, new Double(sum + adder));
      }
    }

    // compute the mean
    for (int i = 0; i < this.attrlist.size(); i++) {
      double sum = ( (Double) meanTuple.get(i)).doubleValue();
      meanTuple.set(i, new Double(sum / cluster.size()));
    }

    return meanTuple;
  }

  /**
   * Output the clustering results.
   */
  public boolean output(String filename) {
    try {
      // check if any of the possible output files exists.
      boolean fileExist = false;
      for (int i = 0; i < this.centroidlist.size(); i++) {
        File outputFile = new File("./" + filename + "_" + (i + 1) + ".rsl");
        if (outputFile.exists()) {
          fileExist = true;
          break;
        }
      }

      if (fileExist) {
        BufferedReader lineReader = new BufferedReader(new InputStreamReader(
            System.in));
        while (true) {
          System.out.print("The filename has existed. Overwrite? (Y/N)");
          String st = lineReader.readLine();
          if (st.equals("Y")) {
            break;
          }
          else if (st.equals("N")) {
            return false;
          }
        }
      } // if

      // Output: 1. Print the centriod and the number of data points in each cluster.
      // 2. Store the actual data points of each cluster in a separate file.
      for (int i = 0; i < this.centroidlist.size(); i++) {
        ArrayList centroid = (ArrayList)this.centroidlist.get(i);
        ArrayList cluster = (ArrayList)this.clusterlist.get(i);

        // 1. Print the centriod and the number of data points in each cluster.
        int size = cluster.size();
        System.out.println("--------  Cluster " + (i + 1) + " (" + size +
                           " data)  --------");
        System.out.print("    Centroid: (");
        for (int j = 0; j < this.attrlist.size(); j++) {
          double value = ( (Double) centroid.get(j)).doubleValue();
          if (j == this.attrlist.size() - 1) {
            System.out.println(this.double2string(value, 2) + ")\n"); // last attribute.
          }
          else {
            System.out.print(this.double2string(value, 2) + ", ");
          }
        }

        // 2. Store the actual data points of each cluster in a separate file.
        File outputFile = new File("./" + filename + "_" + (i + 1) + ".rsl");
        BufferedWriter writer = new BufferedWriter(new FileWriter(outputFile, false));
        StringBuffer buf = new StringBuffer();
        for (int m = 0; m < cluster.size(); m++) {
          ArrayList tuple = (ArrayList)this.valuelist.get( ( (Integer) cluster.
              get(m)).intValue());
          for (int j = 0; j < this.attrlist.size(); j++) {
            double value = ( (Double) tuple.get(j)).doubleValue();
            if (j == this.attrlist.size() - 1) {
              buf.append(this.double2string(value, 2) + "\n"); // last attribute.
            }
            else {
              buf.append(this.double2string(value, 2) + ", "); // last attribute.
            }
          }
        }
        writer.write(buf.toString());
        writer.close();
      } // for
    }
    catch (IOException e) {
      e.printStackTrace();
      return false;
    }
    return true;
  }

  private String double2string(double d, int decimal) {
    String str = Double.toString(d);
    int length = str.indexOf('.') + decimal + 1;
    if (str.length() < length) {
      length = str.length();
    }
    return str.substring(0, length);
  }
}
